-- Only enable for paladins!
local _, PLAYER_CLASS = UnitClass("player")
if PLAYER_CLASS ~= "PALADIN" then return end

Blessed = LibStub("AceAddon-3.0"):NewAddon("Blessed", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceBucket-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Blessed")
local mod = Blessed
local ACD3 = LibStub("AceConfigDialog-3.0")
local LBF = LibStub("LibButtonFacade", true)
local buttonSerial, buttonSubSerial = 1, 1
local db

local GetNumRaidMembers = _G.GetNumRaidMembers 
local GetNumPartyMembers = _G.GetNumPartyMembers
local table, tremove, tinsert = _G.table, _G.tremove, _G.tinsert
local pairs, ipairs = _G.pairs, _G.ipairs
local wipe, select = _G.wipe, _G.select
local UnitIsUnit, UnitAura, UnitClass = _G.UnitIsUnit, _G.UnitAura, _G.UnitClass
local GetSpellInfo, GetSpellLink = _G.GetSpellInfo, _G.GetSpellLink
local GetTime = _G.GetTime

local options = {
	type = "group",
	get = function(info)
		return db[info[#info]]
	end,
	set = function(info, v)
		db[info[#info]] = v
		mod:UpdateLayout()
	end,
	args = {
		scale = {
			name = L["Scale"],
			type = "range",
			min = 0.25,
			max = 4.0,
			step = 0.05,
			bigStep = 0.05
		},
		flip = {
			name = L["Reverse direction"],
			type = "toggle"
		},
		gap = {
			name = L["Button Gap"],
			type = "range",
			min = -5,
			max = 20,
			step = 1,
			bigStep = 1
		}
	}
}

local buffs = {
	-- Not a real buff!
	omni = { 33776 },
	-- It's important that the Lesser and Greater seals have the same ordering
	lesser = {
		20217,		-- Kings
		48932,		-- Might
		48936,		-- Wis
		20911,		-- Sanc
	},
	greater = {
		25898,		-- Kings
		48934,		-- Might
		48938,		-- Wis
		25899,		-- Sanc
	},
	seals = {
		53736,
		20164,
		20165,
		21084,
		31801,
		20166
	},
	auras = {
		19746,
		32223,
		48942,
		48947,
		48945,
		54043,
		48943
	},
	rf = {
		25780
	},
	sense_undead = {
		5502
	}
}

local interestingSpellNames = {}

for listName, list in pairs(buffs) do
	for _, spell in ipairs(list) do
		local name, rank = GetSpellInfo(spell)
		interestingSpellNames[name] = true
	end
end

local buttonPrototype, subButtonPrototype = {}, {}
local buttonPool, subButtonPool, activeButtons = {}, {}, {}
local setAssignments = {}
local raidBuffButtons = {}

local spellNames = setmetatable({}, {__index = function(t, k)
	local name, rank = GetSpellInfo(k)
	local spellName = name -- rank and rank ~= "" and ("%s(%s)"):format(name, rank) or name
	rawset(t, k, spellName)
	return spellName
end})

local forceUpdate
do
	local SCAN_FREQUENCY = 5
	local dummy = CreateFrame("Frame")
	local lastUpdate, lastScan = 0, 0
	function forceUpdate(offset)
		offset = offset or 0.4
		lastUpdate, lastScan = math.max(lastUpdate, SCAN_FREQUENCY - offset), math.max(lastScan, SCAN_FREQUENCY - offset)
	end
	dummy:SetScript("OnUpdate", function(self, elapsed)
		lastUpdate = lastUpdate + elapsed
		lastScan = lastScan + elapsed
		if lastUpdate > 1 then
			local forceScan = false
			if lastScan > SCAN_FREQUENCY then
				lastScan = lastScan - SCAN_FREQUENCY
				forceScan = true
			end
			local unitForNextCast, nextCastSpell, unitsMissingBuffs, unitsOutOfRangeMissingBuffs = nil, nil, 0, 0
			local omni, unit, spell, missing, minTime, buttonMinTime, missingOOR
			for index, button in ipairs(activeButtons) do
				if button.isOmni then
					-- Omni is delayed until the end, where we have the info needed to update it without doing a second full scan.
					omni = button
				elseif forceScan then
					unit, spell, missing, missingOOR, buttonMinTime = button:UpdateUnit()
					unitsMissingBuffs = missing > unitsMissingBuffs and missing or unitsMissingBuffs
					unitsOutOfRangeMissingBuffs = missingOOR > unitsOutOfRangeMissingBuffs and missingOOR or unitsOutOfRangeMissingBuffs
					if (minTime == nil or buttonMinTime < minTime) and buttonMinTime >= 0 then
						unitForNextCast = unit
						nextCastSpell = spell
						minTime = buttonMinTime
					end
				else
					button:UpdateTimer()
				end
				if button.singlePopupHeader:IsVisible() then
					for k, v in ipairs(button.playerButtons) do
						v:UpdateTimer()
					end
				end				
			end
			if omni and forceScan then
				omni:UpdateUnit(unitForNextCast, nextCastSpell, unitsMissingBuffs, unitsOutOfRangeMissingBuffs, minTime)
			end
			lastUpdate = lastUpdate - 1
		end
	end)
end

mod.backdrop = {
	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	insets = {left = 4, top = 4, right = 4, bottom = 4},
	edgeSize = 16,
	tile = true
}

local defaults = {
	profile = {
		scale = 1,
		spellSettings = {},
		singles = {},
		gap = 2
	}
}

local function groupIter(state, index)
	if index < 0 then return end
	local raid, party = GetNumRaidMembers(), GetNumPartyMembers()
	local prefix = raid > 0 and "raid" or "party"
	local unit = prefix .. index
	if UnitExists(unit) then
		return index + 1, unit
	elseif raid == 0 then
		return -1, "player"
	end
end

local function group()
	return groupIter, nil, 1
end

local function getSpellSetting(id, default)
	local specID = GetActiveTalentGroup()
	local val = tonumber(db.spellSettings["spec" .. specID] and db.spellSettings["spec" .. specID][id] or default)
	return val
end

local function setSpellSetting(id, val)
	local specID = GetActiveTalentGroup()
	db.spellSettings["spec" .. specID] = db.spellSettings["spec" .. specID] or {}
	db.spellSettings["spec" .. specID][id] = val
end

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("BlessedDB", defaults)
	db = self.db.profile
	
	self.commPrefix = "BLSS"
	self:RegisterComm(self.commPrefix)
	
	self.lbfGroup = LBF:Group("Blessed", "Buttons")
	self.anchor = CreateFrame("Frame", "BlessedAnchor", UIParent, "SecureHandlerStateTemplate")
	self.anchor:SetPoint("CENTER")
	self.anchor:SetWidth(20)
	self.anchor:SetHeight(20)
	self.anchor:SetMovable(true)
	
	options.args.profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Blessed", options)	
	ACD3:AddToBlizOptions("Blessed")
	ACD3:AddToBlizOptions("Blessed", L["Profiles"], "Blessed", "profiles")
end

function mod:OnEnable()
	self:RunPostCombatAction(self.SetupButtons, self)
	-- self:ScheduleRepeatingTimer("ScanBuffTimers", 3)
	-- self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterBucketEvent("PARTY_MEMBERS_CHANGED", 3, "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	
	self:RegisterEvent("PLAYER_ALIVE")
	forceUpdate(0)
end

local function getNumGroupPaladins()
	local raid, party = GetNumRaidMembers(), GetNumPartyMembers()
	local prefix, num = raid > 0 and "raid" or "party", raid > 0 and raid or party
	local count = 0
	for i = 1, num do
		local _, uc = UnitClass(prefix .. i)
		if uc == "PALADIN" then
			count = count + 1
		end
	end
	return count
end

function mod:PLAYER_ALIVE()
	self:RunPostCombatAction(self.SetupButtons, self)
	self:UnregisterEvent("PLAYER_ALIVE")
	forceUpdate(0)
end

function mod:PARTY_MEMBERS_CHANGED()
	local gnum = GetNumRaidMembers() > 0 and GetNumRaidMembers() or GetNumPartyMembers()
	if gnum ~= self.lastPartySize then
		self:RunPostCombatAction(self.SetRaidBuffButtons, self)
		self:RunPostCombatAction(self.ArrangeButtons, self)
		
		local numPaladins = getNumGroupPaladins()
		
		-- If the number of paladins in the party changes, then start from the beginning, comm talents, and trigger buff assignment.
		if self.lastPaladinsInParty ~= numPaladins then
			self:WipeCachedBlessers()
			
			-- Delay 2 seconds so everyone hasa chance to have wiped their talent set
			self:CancelTimer(self.SendTalentTimer, true)		
			self.SendTalentTimer = self:ScheduleTimer("SendTalents", 2)
			self.lastPaladinsInParty = numPaladins
		else
			-- Just schedule a reassignment
			self:AssignBuffs()
		end
		
		self.lastPartySize = gnum
	end
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(eventName, timestamp, event, _, _, _, _, _, _, spellID, spellName)
	if (event == "SPELL_CAST_SUCCESS" or event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REMOVED") and interestingSpellNames[spellName] then
		forceUpdate()
	end
end

function mod:ACTIVE_TALENT_GROUP_CHANGED()
	self:SendTalents()
	self:SetupButtons()
end

function mod:OnDisable()
	self:RecycleButtons()
end

function mod:RecycleButtons()
	while #activeButtons > 0 do
		local b = tremove(activeButtons)
		b:Free()
	end
	wipe(raidBuffButtons)
end

function mod:SetupButtons()
	self:RecycleButtons()	
	
	self:AddButton("omni", false, nil, buffs.omni)
	self:AddButton("player", true, "PLAYER_AURA", buffs.auras)	
	-- We'll only show the RF button for prot paladins.
	if select(5, GetTalentInfo(2, 7)) > 0 then
		self:AddButton("player", true, "PLAYER_RIGHTEOUS_FURY", buffs.rf)
	end	
	self:AddButton("player", true, "PLAYER_SEAL", buffs.seals)
	
	self:SetRaidBuffButtons()	
	self:UpdateLayout()	
	forceUpdate(0)
	self:ArrangeButtons()
end

function mod:GetDistribution()
	return GetNumRaidMembers() > 0 and "RAID" or "PARTY"
end

function mod:UpdateLayout()
	local f = self.anchor
	local x, y
	x = f:GetLeft()
	y = f:GetBottom()
	-- local delta = self.anchor:GetScale() / db.scale
	-- x = x * delta
	-- y = y * delta
	db.x = x
	db.y = y
	self.anchor:SetScale(db.scale)
	if db.x and db.y then
		self.anchor:ClearAllPoints()
		self.anchor:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", db.x, db.y)
	end
	
	for _, button in ipairs(activeButtons) do
		self:UpdateButtonLayout(button)
	end
end

do
	local leaveCombatActionQueue = {}
	local closures = {}
	function mod:EnqueuePostCombatAction(f, ...)
		if not closures[f] then
			closures[f] = function(...)
				f(...)
			end
		end
		tinsert(leaveCombatActionQueue, 1, closures[f])
	end

	function mod:PLAYER_REGEN_ENABLED()
		while #leaveCombatActionQueue > 0 do
			local f = tremove(leaveCombatActionQueue)
			f(self)
		end
	end
end

do	
	local classes = {}
	local classKeys = {}
	local classUnits = {}
	
	function mod:RunPostCombatAction(f, ...)
		if type(f) ~= "function" then
			error(("Can only queue functions for post-combat execution, arg1 is a %s"):format(type(f)))
		end
		if InCombatLockdown() then		
			self:EnqueuePostCombatAction(f, ...)
		else
			f(...)
		end
	end
	
	local function addRaidUnit(unit)
		local _, uc = UnitClass(unit)
		if uc then
			classes[uc] = (classes[uc] or 0) + 1
			classUnits[uc] = classUnits[uc] or {}
			tinsert(classUnits[uc], unit)
		end
	end
	
	function mod:SetRaidBuffButtons()
		for k, v in pairs(classUnits) do
			wipe(v)
		end
		
		while #raidBuffButtons > 0 do
			local button = tremove(raidBuffButtons)
			button:Free()
		end
		
		wipe(classes)
		wipe(classKeys)
		
		for index, unit in group() do
			addRaidUnit(unit)
		end
		
		for k, v in pairs(classes) do
			tinsert(classKeys, k)
		end
		table.sort(classKeys)
		
		for _, v in ipairs(classKeys) do
			local id = "RAID_BLESSING_" .. v
			local button = self:AddButton(v, false, id, buffs.greater, buffs.lesser)
			for _, unit in ipairs(classUnits[v]) do
				self:AddPlayerButton(button, unit)
			end
			tinsert(raidBuffButtons, button)
		end		
	end
end

function mod:ArrangeButtons()
	local lastButton = self.anchor
	for i = 1, #activeButtons do
		local button = activeButtons[i]
		button:ClearAllPoints()
		button:SetPoint("LEFT", lastButton, "RIGHT", db.gap, 0)
		lastButton = button
	end
end

function mod:AddButton(unitOrClass, isSelfBuff, id, ...)
	local button = tremove(buttonPool)
	if not button then
		button = CreateFrame("Button", "BlessedButton" .. buttonSerial, self.anchor, "SecureActionButtonTemplate SecureHandlerStateTemplate SecureHandlerEnterLeaveTemplate")
		buttonSerial = buttonSerial + 1
		
		button.popupHeader = CreateFrame("Frame", button:GetName() .. "PopupHandler", button, "SecureHandlerAttributeTemplate")
		button.popupHeader:EnableMouse(false)		
		button.popupHeader:SetWidth(38)
		button.popupHeader:SetFrameStrata("DIALOG")
		
		button.singlePopupHeader = CreateFrame("Frame", button:GetName() .. "SinglePopupHandler", button, "SecureHandlerAttributeTemplate")
		button.singlePopupHeader:EnableMouse(false)		
		button.singlePopupHeader:SetWidth(38)
		button.singlePopupHeader:SetHeight(238)
		button.singlePopupHeader:SetFrameStrata("DIALOG")
		button.singlePopupHeader:SetScript("OnShow", function(self)
			for k, v in pairs(self:GetParent().playerButtons) do
				v:UpdateTimer()
			end
		end)
		
		button.spellGroups = {}
		button.subButtons = {}
		button.playerButtons = {}
		
		-- button.singlePopupHeader:SetBackdrop(mod.backdrop)
		button.popupHeader:Hide()
		button.singlePopupHeader:Hide()
		
		button:SetHeight(32)
		button:SetWidth(32)		
		
		button.icon = button:CreateTexture()
		button.icon:SetAllPoints()
		
		button.highlight = button:CreateTexture()
		button:SetHighlightTexture([[Interface\BUTTONS\UI-CheckBox-Highlight]])		
		button.highlight:SetTexture([[Interface\BUTTONS\UI-CheckBox-Highlight]])
		self.lbfGroup:AddButton(button, {Icon = button.icon, Highlight = button.highlight})
		
		button.class = button:CreateTexture(nil, "BACKGROUND")
		button.class:SetWidth(20)
		button.class:SetHeight(20)
		button.class:SetTexture([[Interface\WorldStateFrame\Icons-Classes]])
		
		button.timer = button:CreateFontString(nil, nil, "SystemFont_Outline_Small")
		button.timer:SetText("30:00")
		button.timer:SetPoint("BOTTOM", button, "BOTTOM", 0, 3)
		
		-- Can't set a metatable on secure frames, so just patch in our functionality
		button.Free = buttonPrototype.Free
		button.Cycle = buttonPrototype.Cycle
		button.UpdateTexture = buttonPrototype.UpdateTexture
		button.UpdateTimer = buttonPrototype.UpdateTimer
		button.UpdateUnit = buttonPrototype.UpdateUnit
		button.GetTimeRemaining = buttonPrototype.GetTimeRemaining
		button.SetTimeRemaining = buttonPrototype.SetTimeRemaining
		
		button:SetAttribute("type", "spell")
		button:SetFrameRef("popupHeader", button.popupHeader)
		button:SetFrameRef("singlePopupHeader", button.singlePopupHeader)
		button:SetAttribute("_onenter", [[
			if IsAltKeyDown() then
				local popup = self:GetFrameRef("popupHeader")
				popup:Show()
				popup:RegisterAutoHide()				
			else  -- if not PlayerInCombat() then --if IsControlKeyDown() then
				local popup = self:GetFrameRef("singlePopupHeader")
				popup:Show()
				popup:RegisterAutoHide()
				for i = 1, 20 do
					local f = popup:GetFrameRef("popout" .. i)
					if not f then break end
					local flyout = f:GetFrameRef("flyout")
					popup:AddToAutoHide(flyout)
				end
			end
		]])
		
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
		button:RegisterForDrag("LeftButton")
		
		button:SetScript("OnDragStart", function(self) self:GetParent():StartMoving() end)
		button:SetScript("OnDragStop", function(self)
			local f = self:GetParent()
			f:StopMovingOrSizing()
			db.x = f:GetLeft()
			db.y = f:GetBottom()
		end)
		
		button:EnableMouseWheel(true)
		button.onMousewheel = function(self, dir) self:Cycle(dir) end
		
		button:HookScript("OnEnter", function(self)
			self.showSpell = true
			self:UpdateTexture()
		end)
		
		button:HookScript("OnLeave", function(self)
			self.showSpell = false
			self:UpdateTexture()
		end)
	end
	
	button.id = id
	button.timer:Hide()
	self:UpdateButtonLayout(button)
	
	local l = self.lbfGroup
	
	wipe(button.spellGroups)
	for i = 1, select("#", ...) do
		local spellGroup = select(i, ...)
		tinsert(button.spellGroups, spellGroup)
	end
	
	button.index = 1
	button.lastScan = 0
	local firstGroup = button.spellGroups[1]
	local spellToUse = getSpellSetting(id, firstGroup[1])
	for i = 1, #firstGroup do
		if firstGroup[i] == spellToUse then
			button.index = i
			break
		end
	end
	
	button:Show()
	button.isOmni = unitOrClass == "omni"
	button.isClass = RAID_CLASS_COLORS[unitOrClass] ~= nil
	button.isPlayer = not button.isOmni and not button.isClass
	button.unitOrClass = unitOrClass
	button.showSpell = false
	
	button.class:Hide()
	-- if not button.isClass then
		
	-- elseif button.isClass then
		-- button.class:SetTexCoord(unpack(CLASS_BUTTONS[unitOrClass]))
		-- button.class:Show()
	-- end	
	
	if button.isOmni then
		button:SetScript("OnMousewheel", nil)
		local _, _, icon = GetSpellInfo(button.spellGroups[1][1])
		button.icon:SetTexture(icon)
		SetBindingClick(",", button:GetName(), "LeftButton")
	elseif button.isPlayer then
		button:SetAttribute("unit1", nil)
		button:SetAttribute("unit2", nil)
		button:SetAttribute("unit", button.unitOrGroup)
		button:SetScript("OnMousewheel", button.onMousewheel)
	else
		button:Cycle(0)
		button:SetScript("OnMousewheel", button.onMousewheel)
	end
	
	local firstSpellGroup = button.spellGroups[1]
	if firstSpellGroup then
		for index, spell in ipairs(firstSpellGroup) do
			if GetSpellInfo(GetSpellInfo(spell)) then
				self:AddSubButton(button, index)
			end
		end
	end
	
	tinsert(activeButtons, button)
	return button
end

function mod:UpdateButtonLayout(button)
	button.class:ClearAllPoints()
	button.popupHeader:ClearAllPoints()
	button.singlePopupHeader:ClearAllPoints()
	if db.flip then
		button.class:SetPoint("TOP", button, "BOTTOM", 0, -db.gap)
		button.popupHeader:SetPoint("BOTTOM", button, "BOTTOM")
		button.singlePopupHeader:SetPoint("BOTTOM", button, "BOTTOM")
	else
		button.class:SetPoint("BOTTOM", button, "TOP", 0, db.gap)
		button.popupHeader:SetPoint("TOP", button, "TOP")
		button.singlePopupHeader:SetPoint("TOP", button, "TOP")
	end
	if button.subButtons then
		for index, sub in ipairs(button.subButtons) do
			sub:UpdateLayout(index)
		end
	end
end

local function clickSubButton(self)
	if self.isSubButton then	
		local parent = self.parent
		
		local link = GetSpellLink(parent:GetAttribute("spell1"))
		if link then
			local id = tonumber(link:match("Hspell:(%d+)"))
			for i = 1, #parent.spellGroups[1] do
				if parent.spellGroups[1][i] == id then
					parent.index = i
					if parent.id then
						setSpellSetting(parent.id, parent.spellGroups[1][i])
					end				
					break
				end
			end
			forceUpdate(0)			
		end
	end
end

local function createSubButton(self)
	local button = tremove(subButtonPool)
	if not button then
		button = CreateFrame("Button", "BlessedSubButton" .. buttonSubSerial, mod.anchor, "SecureActionButtonTemplate SecureHandlerMouseUpDownTemplate SecureHandlerAttributeTemplate SecureHandlerEnterLeaveTemplate")

		button.text = button:CreateFontString(nil, nil, "SystemFont_Outline_Small")
		button.text:SetPoint("LEFT", button, "RIGHT", 2, 0)

		button.timer = button:CreateFontString(nil, nil, "SystemFont_Outline_Small")
		button.timer:SetPoint("RIGHT", button, "LEFT", -2, 0)
		
		buttonSubSerial = buttonSubSerial + 1
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp")
		button:SetHeight(32)
		button:SetWidth(32)
		
		button.flyoutHeader = CreateFrame("Frame", button:GetName() .. "flyoutHeader", button, "SecureHandlerAttributeTemplate")
		button.flyoutHeader:EnableMouse(false)		
		button.flyoutHeader:SetHeight(38)
		button.flyoutHeader:SetFrameStrata("DIALOG")
		button.flyoutHeader:Hide()
		button.flyoutHeader:SetPoint("LEFT", button, "LEFT")
		button.flyoutHeader:SetScript("OnShow", function(self) self:GetParent().text:Hide() end)
		button.flyoutHeader:SetScript("OnHide", function(self) self:GetParent().text:Show() end)
		button:SetFrameRef("flyout", button.flyoutHeader)
		button.Free = subButtonPrototype.Free
		button.UpdateTimer = subButtonPrototype.UpdateTimer
		button.UpdateLayout = subButtonPrototype.UpdateLayout		
		
		button.subButtons = {}

		button.icon = button:CreateTexture()
		button.icon:SetAllPoints()

		button.highlight = button:CreateTexture()
		button:SetHighlightTexture([[Interface\BUTTONS\UI-CheckBox-Highlight.blp]])		
		button.highlight:SetTexture([[Interface\BUTTONS\UI-CheckBox-Highlight.blp]])
		
		button:HookScript("OnClick", clickSubButton)
		
		self.lbfGroup:AddButton(button, {Icon = button.icon, Highlight = button.highlight})
	end	
	button:Show()
	button.timer:SetText("")
	button.text:Hide()
	button:SetAttribute("type", nil)
	button:SetAttribute("unit", nil)
	button:SetAttribute("spell", nil)
	button:SetAttribute("_onenter", nil)
	button:SetAttribute("_onmouseup", nil)
	button:SetAttribute("_onattributechanged", nil)
	button:SetScript("OnAttributeChanged", nil)
	
	button:UnregisterEvent("UNIT_PORTRAIT_UPDATE")
	button:SetScript("OnEvent", nil)
	
	button.type = nil
	button.container = nil
	button.parent = nil
	button.isSubButton = false
	return button
end

local function singleCastChanged(self, key, val)
	if key == "spell" then
		local unit = self:GetAttribute("unit")
		if not unit then return end
		local name = UnitName(unit)
		if not name then return end
		
		local spell = val
		if spell then
			local link = GetSpellLink(spell)
			if link then
				local id = tonumber(link:match("Hspell:(%d+)"))
				db.singles[name] = id
				forceUpdate(0)
				return
			end
		end
		db.singles[name] = nil
		forceUpdate(0)
	end
end

local function createSingleButton(self, button, unit, spell)
	local single = createSubButton(self)
	single:SetParent(button.flyoutHeader)		
	single.type = "horizontal"
	single.parent = button
	single.container = button.subButtons
	tinsert(single.container, single)
	single:ClearAllPoints()
	
	if spell then
		local _, _, icon = GetSpellInfo(spell)
		single.icon:SetTexture(icon)
		single:SetAttribute("setSpell", spellNames[spell])
	else
		single.icon:SetTexture([[Interface\Icons\Ability_Vehicle_PlayerLoaded]])
		single.icon:SetTexCoord(0.774, 0.324, 0.067, 0.489)
		single:SetAttribute("setSpell", nil)
	end
	
	single:SetAttribute("attribute-frame", button)
	single:SetAttribute("attribute-name", "spell")
	single:SetAttribute("attribute-value", spell and spellNames[spell] or nil)
	single:SetAttribute("type", "attribute")
	
	single:UpdateLayout()
end

do
	local function updatePortrait(self, unit)
		local sunit = self:GetAttribute("unit")
		if sunit and UnitIsUnit(unit, sunit) then
			SetPortraitTexture(self.icon, unit)
		end
	end

	local function playerButtonOnEvent(self, event, ...)
		if event == "UNIT_PORTRAIT_UPDATE" then
			updatePortrait(self, ...)
		end
	end

	function mod:AddPlayerButton(parent, unit)
		local button = createSubButton(self)
		button.parent = parent
		button:SetParent(parent.singlePopupHeader)	
		button:SetFrameRef("header", parent.singlePopupHeader)
		button:SetScript("OnAttributeChanged", singleCastChanged)
		button.text:Show()	
		button.text:SetText(UnitName(unit))
		
		button:RegisterEvent("UNIT_PORTRAIT_UPDATE")	
		button:SetScript("OnEvent", playerButtonOnEvent)
		SetPortraitTexture(button.icon, unit)
		
		createSingleButton(self, button, unit, nil)
		for _, spell in pairs(parent.spellGroups[#parent.spellGroups]) do
			createSingleButton(self, button, unit, spell)
		end
			
		button.type = "vertical"
		button.container = parent.playerButtons
		
		button:SetAttribute("_onenter", [[
			local popup = self:GetFrameRef("flyout")
			popup:Show()
			popup:RegisterAutoHide(0.01)
		]])
		button:SetAttribute("_onattributechanged", [[
			self:SetAttribute("saveSpellToDB", true)
			self:GetFrameRef("flyout"):Hide()
		]])
		button:SetAttribute("type", "spell")
		button:SetAttribute("unit", UnitName(unit))
		if db.singles[UnitName(unit)] then
			local name = GetSpellInfo(db.singles[UnitName(unit)])
			button:SetAttribute("spell", name)
		end
		
		
		tinsert(button.container, button)
		parent.singlePopupHeader:SetFrameRef("popout" .. #button.container, button)
		button:UpdateLayout()
	end
end

function mod:AddSubButton(parent, index)
	local button = createSubButton(self)
	
	button.parent = parent	
	button:SetParent(parent.popupHeader)
	button.type = "vertical"
	button.isSubButton = true
	button.container = parent.subButtons
	
	local name, _, icon = GetSpellInfo(parent.spellGroups[1][index])
	button.icon:SetTexture(icon)
	button.text:SetText(name)
	button.text:Show()	

	button:SetFrameRef("parent", parent)
	button:SetFrameRef("header", parent.popupHeader)
	for i = 1, #parent.spellGroups do
		local spell = parent.spellGroups[i][index]
		local spellName = spellNames[spell]
	end
	button:SetAttribute("_onmouseup", [[
		local parent = self:GetFrameRef("parent")
		for i = 1, 20 do
			if spellName then
				parent:SetAttribute("spell" .. i, spellName)
			else
				break
			end
		end
		self:GetFrameRef("header"):Hide()
	]])
	
	tinsert(button.container, button)
	button:UpdateLayout()
end

function buttonPrototype:Free()
	tinsert(buttonPool, self)
	for i, v in ipairs(activeButtons) do
		if v == self then
			tremove(activeButtons, i)
			break
		end
	end
	while #self.subButtons > 0 do
		tremove(self.subButtons):Free()
	end
	while #self.playerButtons > 0 do
		tremove(self.playerButtons):Free()
	end
	self:Hide()
	self:ClearAllPoints()
end

function subButtonPrototype:UpdateLayout(index)
	local parent = self.parent
	self:ClearAllPoints()
	index = index or #self.container
	offset = (self:GetHeight() + db.gap) * index
	if self.type == "horizontal" then
		self:SetPoint("LEFT", parent, "LEFT", offset, 0)
		self:GetParent():SetWidth((#self.container + 1.5) * (self:GetHeight() + db.gap))
	else
		if db.flip then
			self:SetPoint("BOTTOM", parent, "BOTTOM", 0, offset)
		else
			self:SetPoint("TOP", parent, "TOP", 0, -offset)
		end
		self:GetParent():SetHeight((#self.container + 1.5) * (self:GetHeight() + db.gap))
	end
end

function subButtonPrototype:Free()
	while #self.subButtons > 0 do
		tremove(self.subButtons):Free()
	end	
	tinsert(subButtonPool, self)
	self:Hide()
	self:ClearAllPoints()
end

local function expiryText(unit, spell)
	local spellName, _, icon = GetSpellInfo(spell)
	local spellIcon = ("|T%s:38|t"):format(icon)
	local expiresAt = select(7, UnitAura(unit, spellName))
	if not expiresAt or expiresAt - GetTime() <= 0 then
		return ("|cffee0000--:--|r %s"):format(spellIcon), false
	else
		local timeLeft = expiresAt - GetTime()
		local minutes = math.ceil(timeLeft / 60)
		local seconds = timeLeft % 60
		if minutes > 5 then
			text = ("|cff00ee00%2dm|r"):format(minutes)
		else
			text = ("|cffffff00%2d:%02d|r"):format(minutes - 1, seconds)
		end
		return ("%s %s"):format(text, spellIcon), true
	end
end

function subButtonPrototype:UpdateTimer()
	local spell = self:GetAttribute("spell")
	local unit = self:GetAttribute("unit")
	local timerText
	if spell then
		timerText = expiryText(unit, spell)
	else
		for idx, group in ipairs(self.parent.spellGroups) do
			local spell = group[self.parent.index]
			if spell then
				local text, hasTime = expiryText(unit, spell)
				timerText = timerText and (hasTime and text or timerText) or text
				if hasTime then break end
			end
		end
	end
	self.timer:SetText(timerText)
end

function buttonPrototype:Cycle(direction)
	local group = self.spellGroups[1]
	if #group == 1 then
		forceUpdate(0)
		return
	end
	
	local newSpell	
	for i = 1, #group do
		self.index = self.index + direction
		if group[self.index] == nil then
			if direction == 1 then
				self.index = 1
			else
				self.index = #group
			end
		end
		newSpell = group[self.index]
		
		-- If the new spell is in our spellbook, then we'll stay here!
		if GetSpellInfo(GetSpellInfo(newSpell)) then
			break
		end
	end
	if self.id then
		setSpellSetting(self.id, group[self.index])
	end
	forceUpdate(0)
end

do
	local spellMaxDuration = setmetatable({}, {__index = function(t, k)
		local link = GetSpellLink(k)
		if not link then
			rawset(t, k, 0)
			return 0
		end
		local spellID = tonumber(link:match("Hspell:(%d+)"))
		for _, v in ipairs(buffs.lesser) do
			if spellID == v then
				rawset(t, k, 600)
				return 600
			end
		end
		for _, v in ipairs(buffs.greater) do
			if spellID == v then
				rawset(t, k, 1800)
				return 1800
			end
		end
		rawset(t, k, 0)
		return 0
	end})
	
	local lesserToGreater = setmetatable({}, {__index = function(t, k)
		for index, buffID in ipairs(buffs.lesser) do
			local n = GetSpellInfo(buffID)
			if n == k then
				local gn = GetSpellInfo(buffs.greater[index])
				rawset(t, k, gn)
				return gn
			end
		end
		
		rawset(t, k, false)
		return false
	end})
	
	local function nextCastInfoForUnit(self, unit, useSingleOverride, spellIDOverride)
		local inRange, spellToCast, hasBuff, maxTimeLeft
		for i = 1, #self.spellGroups do
			local spellID = spellIDOverride or (useSingleOverride and db.singles[UnitName(unit)]) or self.spellGroups[1][self.index]
			local spellName = GetSpellInfo(spellID)
			if not spellName then break end
			
			-- First, check to see if someone has cast one of our one-shot lessers on this target. If so, we will just ignore it.
			local greaterSpell = lesserToGreater[spellName]
			local greaterExpiresAt = greaterSpell and select(7, UnitAura(unit, greaterSpell)) or nil
			local expiresAt = greaterExpiresAt
			
			if greaterExpiresAt then		-- Someone else cast our lesser on this target. We don't need to buff them.
				hasBuff = true
				spellToCast = nil
			else
				expiresAt = select(7, UnitAura(unit, spellName))
				hasBuff = hasBuff ~= true and expiresAt ~= nil or hasBuff
				spellToCast = spellToCast or spellName
			end
			
			expiresAt = math.max((expiresAt or 0), 0)
			
			-- This is a hack for buffs without a duration
			if expiresAt == 0 and hasBuff then expiresAt = -1 end			
			
			maxTimeLeft = (maxTimeLeft == nil or expiresAt > maxTimeLeft) and expiresAt or maxTimeLeft
			buffTimeLeft = (buffTimeLeft == nil or expiresAt < buffTimeLeft) and expiresAt or buffTimeLeft
			inRange = inRange ~= true and IsSpellInRange(spellName, unit) == 1 or UnitIsUnit("player", unit) or inRange
		end
		
		return hasBuff, inRange, spellToCast, maxTimeLeft
	end
	
	local HUGE_BUFF_TIMER = 99999999
	local classAssignments = setmetatable({}, {__index = function(t, k)
		for _, button in ipairs(activeButtons) do
			if button.isClass and button.unitOrClass == k then
				local spell = button.spellGroups[1][button.index]
				rawset(t, k, spell)
				return spell
			end
		end
	end})
	
	function buttonPrototype:UpdateUnit(omniUnit, omniSpell, omniInRangeMissingBuffs, omniOutOfRangeMissingBuffs, omniTime)
		local minTimer = HUGE_BUFF_TIMER
		local unitsInRangeMissingBuffs, unitsOutOfRangeMissingBuffs = 0, 0
		local castUnit, newSpell = nil, nil
		
		if self.isPlayer then
			local hasBuff, inRange, spellToCast, maxTimeLeft = nextCastInfoForUnit(self, self.unitOrClass)
			unitsInRangeMissingBuffs = hasBuff and 0 or 1
			minTimer = maxTimeLeft
			castUnit = self.unitOrClass
			newSpell = spellToCast
		elseif self.isOmni then
			castUnit = omniUnit
			newSpell = omniSpell
			minTimer = omniTime
			unitsInRangeMissingBuffs = omniInRangeMissingBuffs
			unitsOutOfRangeMissingBuffs = omniOutOfRangeMissingBuffs
		elseif self.isClass then
			for _, playerButton in ipairs(self.playerButtons) do
				local unit = playerButton:GetAttribute("unit")
				local hasBuff, inRange, spellToCast, maxTimeLeft = nextCastInfoForUnit(self, unit, true)
				unitsInRangeMissingBuffs = unitsInRangeMissingBuffs + (inRange and not hasBuff and 1 or 0)
				unitsOutOfRangeMissingBuffs = unitsOutOfRangeMissingBuffs + (not inRange and 1 or 0)
				if minTimer > maxTimeLeft and inRange then
					newSpell = spellToCast
					minTimer = maxTimeLeft
					castUnit = unit
				end
			end
		else
			print("|cffffff00[Blessed]|r |cffff0000ERROR|r: Updating for unknown button type")
			return
		end
		
		if castUnit and not InCombatLockdown() then
			self:SetAttribute("unit", castUnit)
			self:SetAttribute("type", "spell")
			self:SetAttribute("spell", newSpell)
		end
		
		self:SetTimeRemaining(minTimer == HUGE_BUFF_TIMER and 0 or minTimer)
		
		local r, g, b, a = 1, 1, 1, 1
		if unitsOutOfRangeMissingBuffs > 0 and unitsInRangeMissingBuffs == 0 then
			r, g, b = 0.3, 0.3, 0.3
		end
		if unitsInRangeMissingBuffs > 0 then
			r, g, b = 0.85, 0.15, 0.15
		end
		self.icon:SetVertexColor(r, g, b, a)
		self:UpdateTexture()
		
		return castUnit, newSpell, unitsInRangeMissingBuffs, unitsOutOfRangeMissingBuffs, minTimer == HUGE_BUFF_TIMER and 0 or minTimer
	end
end

function buttonPrototype:SetTimeRemaining(t)
	self.expiresAt = t
	self:UpdateTimer()
end

function buttonPrototype:GetTimeRemaining()
	return (self.expiresAt or 0) - GetTime()
end

function buttonPrototype:UpdateTimer()	
	local timeLeft = self:GetTimeRemaining()
	local text
	if timeLeft <= 0 then
		if self.running then
			self.running = false
			forceUpdate()
		end
		text = ""
	elseif timeLeft then
		self.running = true
		local minutes = math.ceil(timeLeft / 60)
		local seconds = timeLeft % 60
		if minutes > 5 then
			text = ("%2dm"):format(minutes)
		else
			text = ("%2d:%02d"):format(minutes - 1, seconds)
		end
	else
		text = ""
	end
	self.timer:SetText(text)
	self.timer:Show()
end

function buttonPrototype:UpdateTexture()
	if self.isOmni then
		self.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
		return
	end
	if self.isClass and not self.showSpell then
		self.icon:SetTexture([[Interface\GLUES\CharacterCreate\UI-CharacterCreate-Classes.blp]])
		local a, b, c, d = unpack(CLASS_BUTTONS[self.unitOrClass])
		local delta = (b - a) * 0.08
		a = a + delta
		b = b - delta
		c = c + delta
		d = d - delta
		self.icon:SetTexCoord(a, b, c, d)
	else
		if not self.spellGroups[1] then return end
		local spell = self.spellGroups[1][self.index]
		if not spell then return end
		local name, rank, icon = GetSpellInfo(spell)
		if not name then return end
		self.icon:SetTexture(icon)
		self.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	end
end

-- TODO: Use this whenever someone changes greaters
-- Make sure that sub buttons are updated appropriately, since they might have their singles yanked out from underneath them.
function buttonPrototype:ClearSinglesForButton()
	for _, b in ipairs(self.playerButtons) do
		local u = b:GetAttribute("unit")
		if u then
			local n = UnitName(u)
			if n then
				db.singles[n] = nil
			end
		end
	end
end